@@ -6,7 +6,7 @@ use wasmer::Val;
 use cosmwasm_std::{ContractResult, Env, MessageInfo, QueryResponse, Reply, Response};
 #[cfg(feature = "stargate")]
 use cosmwasm_std::{
-    IbcAcknowledgement, IbcBasicResponse, IbcChannel, IbcPacket, IbcReceiveResponse,
+    IbcAcknowledgementWithPacket, IbcBasicResponse, IbcChannel, IbcPacket, IbcReceiveResponse,
 };
 
 use crate::backend::{BackendApi, Querier, Storage};
@@ -15,36 +15,85 @@ use crate::errors::{VmError, VmResult};
 use crate::instance::Instance;
 use crate::serde::{from_slice, to_vec};
 
-/// Max length (in bytes) of the result data from an instantiate call.
-const RESULT_LIMIT_INSTANTIATE: usize = 100_000;
-/// Max length (in bytes) of the result data from an execute call.
-const RESULT_LIMIT_EXECUTE: usize = 100_000;
-/// Max length (in bytes) of the result data from a migrate call.
-const RESULT_LIMIT_MIGRATE: usize = 100_000;
-/// Max length (in bytes) of the result data from a sudo call.
-const RESULT_LIMIT_SUDO: usize = 100_000;
-/// Max length (in bytes) of the result data from a reply call.
-const RESULT_LIMIT_REPLY: usize = 100_000;
-/// Max length (in bytes) of the result data from a query call.
-const RESULT_LIMIT_QUERY: usize = 100_000;
-/// Max length (in bytes) of the result data from a ibc_channel_open call.
-#[cfg(feature = "stargate")]
-const RESULT_LIMIT_IBC_CHANNEL_OPEN: usize = 100_000;
-/// Max length (in bytes) of the result data from a ibc_channel_connect call.
-#[cfg(feature = "stargate")]
-const RESULT_LIMIT_IBC_CHANNEL_CONNECT: usize = 100_000;
-/// Max length (in bytes) of the result data from a ibc_channel_close call.
-#[cfg(feature = "stargate")]
-const RESULT_LIMIT_IBC_CHANNEL_CLOSE: usize = 100_000;
-/// Max length (in bytes) of the result data from a ibc_packet_receive call.
-#[cfg(feature = "stargate")]
-const RESULT_LIMIT_IBC_PACKET_RECEIVE: usize = 100_000;
-/// Max length (in bytes) of the result data from a ibc_packet_ack call.
-#[cfg(feature = "stargate")]
-const RESULT_LIMIT_IBC_PACKET_ACK: usize = 100_000;
-/// Max length (in bytes) of the result data from a ibc_packet_timeout call.
-#[cfg(feature = "stargate")]
-const RESULT_LIMIT_IBC_PACKET_TIMEOUT: usize = 100_000;
+/// The limits in here protect the host from allocating an unreasonable amount of memory
+/// and copying an unreasonable amount of data.
+///
+/// A JSON deserializer would want to set the limit to a much smaller value because
+/// deserializing JSON is more expensive. As a consequence, any sane contract should hit
+/// the deserializer limit before the read limit.
+mod read_limits {
+    /// A mibi (mega binary)
+    const MI: usize = 1024 * 1024;
+    /// Max length (in bytes) of the result data from an instantiate call.
+    pub const RESULT_INSTANTIATE: usize = 64 * MI;
+    /// Max length (in bytes) of the result data from an execute call.
+    pub const RESULT_EXECUTE: usize = 64 * MI;
+    /// Max length (in bytes) of the result data from a migrate call.
+    pub const RESULT_MIGRATE: usize = 64 * MI;
+    /// Max length (in bytes) of the result data from a sudo call.
+    pub const RESULT_SUDO: usize = 64 * MI;
+    /// Max length (in bytes) of the result data from a reply call.
+    pub const RESULT_REPLY: usize = 64 * MI;
+    /// Max length (in bytes) of the result data from a query call.
+    pub const RESULT_QUERY: usize = 64 * MI;
+    /// Max length (in bytes) of the result data from a ibc_channel_open call.
+    #[cfg(feature = "stargate")]
+    pub const RESULT_IBC_CHANNEL_OPEN: usize = 64 * MI;
+    /// Max length (in bytes) of the result data from a ibc_channel_connect call.
+    #[cfg(feature = "stargate")]
+    pub const RESULT_IBC_CHANNEL_CONNECT: usize = 64 * MI;
+    /// Max length (in bytes) of the result data from a ibc_channel_close call.
+    #[cfg(feature = "stargate")]
+    pub const RESULT_IBC_CHANNEL_CLOSE: usize = 64 * MI;
+    /// Max length (in bytes) of the result data from a ibc_packet_receive call.
+    #[cfg(feature = "stargate")]
+    pub const RESULT_IBC_PACKET_RECEIVE: usize = 64 * MI;
+    /// Max length (in bytes) of the result data from a ibc_packet_ack call.
+    #[cfg(feature = "stargate")]
+    pub const RESULT_IBC_PACKET_ACK: usize = 64 * MI;
+    /// Max length (in bytes) of the result data from a ibc_packet_timeout call.
+    #[cfg(feature = "stargate")]
+    pub const RESULT_IBC_PACKET_TIMEOUT: usize = 64 * MI;
+}
+
+/// The limits for the JSON deserialization.
+///
+/// Those limits are not used when the Rust JSON deserializer is bypassed by using the
+/// public `call_*_raw` functions directly.
+mod deserialization_limits {
+    /// A kibi (kilo binary)
+    const KI: usize = 1024;
+    /// Max length (in bytes) of the result data from an instantiate call.
+    pub const RESULT_INSTANTIATE: usize = 256 * KI;
+    /// Max length (in bytes) of the result data from an execute call.
+    pub const RESULT_EXECUTE: usize = 256 * KI;
+    /// Max length (in bytes) of the result data from a migrate call.
+    pub const RESULT_MIGRATE: usize = 256 * KI;
+    /// Max length (in bytes) of the result data from a sudo call.
+    pub const RESULT_SUDO: usize = 256 * KI;
+    /// Max length (in bytes) of the result data from a reply call.
+    pub const RESULT_REPLY: usize = 256 * KI;
+    /// Max length (in bytes) of the result data from a query call.
+    pub const RESULT_QUERY: usize = 256 * KI;
+    /// Max length (in bytes) of the result data from a ibc_channel_open call.
+    #[cfg(feature = "stargate")]
+    pub const RESULT_IBC_CHANNEL_OPEN: usize = 256 * KI;
+    /// Max length (in bytes) of the result data from a ibc_channel_connect call.
+    #[cfg(feature = "stargate")]
+    pub const RESULT_IBC_CHANNEL_CONNECT: usize = 256 * KI;
+    /// Max length (in bytes) of the result data from a ibc_channel_close call.
+    #[cfg(feature = "stargate")]
+    pub const RESULT_IBC_CHANNEL_CLOSE: usize = 256 * KI;
+    /// Max length (in bytes) of the result data from a ibc_packet_receive call.
+    #[cfg(feature = "stargate")]
+    pub const RESULT_IBC_PACKET_RECEIVE: usize = 256 * KI;
+    /// Max length (in bytes) of the result data from a ibc_packet_ack call.
+    #[cfg(feature = "stargate")]
+    pub const RESULT_IBC_PACKET_ACK: usize = 256 * KI;
+    /// Max length (in bytes) of the result data from a ibc_packet_timeout call.
+    #[cfg(feature = "stargate")]
+    pub const RESULT_IBC_PACKET_TIMEOUT: usize = 256 * KI;
+}
 
 pub fn call_instantiate<A, S, Q, U>(
     instance: &mut Instance<A, S, Q>,
@@ -61,7 +110,8 @@ where
     let env = to_vec(env)?;
     let info = to_vec(info)?;
     let data = call_instantiate_raw(instance, &env, &info, msg)?;
-    let result: ContractResult<Response<U>> = from_slice(&data)?;
+    let result: ContractResult<Response<U>> =
+        from_slice(&data, deserialization_limits::RESULT_INSTANTIATE)?;
     Ok(result)
 }
 
@@ -80,7 +130,8 @@ where
     let env = to_vec(env)?;
     let info = to_vec(info)?;
     let data = call_execute_raw(instance, &env, &info, msg)?;
-    let result: ContractResult<Response<U>> = from_slice(&data)?;
+    let result: ContractResult<Response<U>> =
+        from_slice(&data, deserialization_limits::RESULT_EXECUTE)?;
     Ok(result)
 }
 
@@ -97,7 +148,8 @@ where
 {
     let env = to_vec(env)?;
     let data = call_migrate_raw(instance, &env, msg)?;
-    let result: ContractResult<Response<U>> = from_slice(&data)?;
+    let result: ContractResult<Response<U>> =
+        from_slice(&data, deserialization_limits::RESULT_MIGRATE)?;
     Ok(result)
 }
 
@@ -114,7 +166,8 @@ where
 {
     let env = to_vec(env)?;
     let data = call_sudo_raw(instance, &env, msg)?;
-    let result: ContractResult<Response<U>> = from_slice(&data)?;
+    let result: ContractResult<Response<U>> =
+        from_slice(&data, deserialization_limits::RESULT_SUDO)?;
     Ok(result)
 }
 
@@ -132,7 +185,8 @@ where
     let env = to_vec(env)?;
     let msg = to_vec(msg)?;
     let data = call_reply_raw(instance, &env, &msg)?;
-    let result: ContractResult<Response<U>> = from_slice(&data)?;
+    let result: ContractResult<Response<U>> =
+        from_slice(&data, deserialization_limits::RESULT_REPLY)?;
     Ok(result)
 }
 
@@ -148,7 +202,8 @@ where
 {
     let env = to_vec(env)?;
     let data = call_query_raw(instance, &env, msg)?;
-    let result: ContractResult<QueryResponse> = from_slice(&data)?;
+    let result: ContractResult<QueryResponse> =
+        from_slice(&data, deserialization_limits::RESULT_QUERY)?;
     // Ensure query response is valid JSON
     if let ContractResult::Ok(binary_response) = &result {
         serde_json::from_slice::<serde_json::Value>(binary_response.as_slice()).map_err(|e| {
@@ -173,7 +228,8 @@ where
     let env = to_vec(env)?;
     let channel = to_vec(channel)?;
     let data = call_ibc_channel_open_raw(instance, &env, &channel)?;
-    let result: ContractResult<()> = from_slice(&data)?;
+    let result: ContractResult<()> =
+        from_slice(&data, deserialization_limits::RESULT_IBC_CHANNEL_OPEN)?;
     Ok(result)
 }
 
@@ -192,7 +248,7 @@ where
     let env = to_vec(env)?;
     let channel = to_vec(channel)?;
     let data = call_ibc_channel_connect_raw(instance, &env, &channel)?;
-    let result = from_slice(&data)?;
+    let result = from_slice(&data, deserialization_limits::RESULT_IBC_CHANNEL_CONNECT)?;
     Ok(result)
 }
 
@@ -211,7 +267,7 @@ where
     let env = to_vec(env)?;
     let channel = to_vec(channel)?;
     let data = call_ibc_channel_close_raw(instance, &env, &channel)?;
-    let result = from_slice(&data)?;
+    let result = from_slice(&data, deserialization_limits::RESULT_IBC_CHANNEL_CLOSE)?;
     Ok(result)
 }
 
@@ -230,7 +286,7 @@ where
     let env = to_vec(env)?;
     let packet = to_vec(packet)?;
     let data = call_ibc_packet_receive_raw(instance, &env, &packet)?;
-    let result = from_slice(&data)?;
+    let result = from_slice(&data, deserialization_limits::RESULT_IBC_PACKET_RECEIVE)?;
     Ok(result)
 }
 
@@ -238,7 +294,7 @@ where
 pub fn call_ibc_packet_ack<A, S, Q, U>(
     instance: &mut Instance<A, S, Q>,
     env: &Env,
-    ack: &IbcAcknowledgement,
+    ack: &IbcAcknowledgementWithPacket,
 ) -> VmResult<ContractResult<IbcBasicResponse<U>>>
 where
     A: BackendApi + 'static,
@@ -249,7 +305,7 @@ where
     let env = to_vec(env)?;
     let ack = to_vec(ack)?;
     let data = call_ibc_packet_ack_raw(instance, &env, &ack)?;
-    let result = from_slice(&data)?;
+    let result = from_slice(&data, deserialization_limits::RESULT_IBC_PACKET_ACK)?;
     Ok(result)
 }
 
@@ -268,7 +324,7 @@ where
     let env = to_vec(env)?;
     let packet = to_vec(packet)?;
     let data = call_ibc_packet_timeout_raw(instance, &env, &packet)?;
-    let result = from_slice(&data)?;
+    let result = from_slice(&data, deserialization_limits::RESULT_IBC_PACKET_TIMEOUT)?;
     Ok(result)
 }
 
@@ -290,7 +346,7 @@ where
         instance,
         "instantiate",
         &[env, info, msg],
-        RESULT_LIMIT_INSTANTIATE,
+        read_limits::RESULT_INSTANTIATE,
     )
 }
 
@@ -308,7 +364,12 @@ where
     Q: Querier + 'static,
 {
     instance.set_storage_readonly(false);
-    call_raw(instance, "execute", &[env, info, msg], RESULT_LIMIT_EXECUTE)
+    call_raw(
+        instance,
+        "execute",
+        &[env, info, msg],
+        read_limits::RESULT_EXECUTE,
+    )
 }
 
 /// Calls Wasm export "migrate" and returns raw data from the contract.
@@ -324,7 +385,12 @@ where
     Q: Querier + 'static,
 {
     instance.set_storage_readonly(false);
-    call_raw(instance, "migrate", &[env, msg], RESULT_LIMIT_MIGRATE)
+    call_raw(
+        instance,
+        "migrate",
+        &[env, msg],
+        read_limits::RESULT_MIGRATE,
+    )
 }
 
 /// Calls Wasm export "sudo" and returns raw data from the contract.
@@ -340,7 +406,7 @@ where
     Q: Querier + 'static,
 {
     instance.set_storage_readonly(false);
-    call_raw(instance, "sudo", &[env, msg], RESULT_LIMIT_SUDO)
+    call_raw(instance, "sudo", &[env, msg], read_limits::RESULT_SUDO)
 }
 
 /// Calls Wasm export "reply" and returns raw data from the contract.
@@ -356,7 +422,7 @@ where
     Q: Querier + 'static,
 {
     instance.set_storage_readonly(false);
-    call_raw(instance, "reply", &[env, msg], RESULT_LIMIT_REPLY)
+    call_raw(instance, "reply", &[env, msg], read_limits::RESULT_REPLY)
 }
 
 /// Calls Wasm export "query" and returns raw data from the contract.
@@ -372,7 +438,7 @@ where
     Q: Querier + 'static,
 {
     instance.set_storage_readonly(true);
-    call_raw(instance, "query", &[env, msg], RESULT_LIMIT_QUERY)
+    call_raw(instance, "query", &[env, msg], read_limits::RESULT_QUERY)
 }
 
 #[cfg(feature = "stargate")]
@@ -391,7 +457,7 @@ where
         instance,
         "ibc_channel_open",
         &[env, channel],
-        RESULT_LIMIT_IBC_CHANNEL_OPEN,
+        read_limits::RESULT_IBC_CHANNEL_OPEN,
     )
 }
 
@@ -411,7 +477,7 @@ where
         instance,
         "ibc_channel_connect",
         &[env, channel],
-        RESULT_LIMIT_IBC_CHANNEL_CONNECT,
+        read_limits::RESULT_IBC_CHANNEL_CONNECT,
     )
 }
 
@@ -431,7 +497,7 @@ where
         instance,
         "ibc_channel_close",
         &[env, channel],
-        RESULT_LIMIT_IBC_CHANNEL_CLOSE,
+        read_limits::RESULT_IBC_CHANNEL_CLOSE,
     )
 }
 
@@ -451,7 +517,7 @@ where
         instance,
         "ibc_packet_receive",
         &[env, packet],
-        RESULT_LIMIT_IBC_PACKET_RECEIVE,
+        read_limits::RESULT_IBC_PACKET_RECEIVE,
     )
 }
 
@@ -471,7 +537,7 @@ where
         instance,
         "ibc_packet_ack",
         &[env, ack],
-        RESULT_LIMIT_IBC_PACKET_ACK,
+        read_limits::RESULT_IBC_PACKET_ACK,
     )
 }
 
@@ -491,7 +557,7 @@ where
         instance,
         "ibc_packet_timeout",
         &[env, packet],
-        RESULT_LIMIT_IBC_PACKET_TIMEOUT,
+        read_limits::RESULT_IBC_PACKET_TIMEOUT,
     )
 }
 
@@ -612,7 +678,10 @@ mod tests {
             mock_env, mock_info, mock_instance, MockApi, MockQuerier, MockStorage,
         };
         use cosmwasm_std::testing::{mock_ibc_channel, mock_ibc_packet_ack};
-        use cosmwasm_std::{attr, Empty, Event, IbcOrder, Reply, SubcallResponse};
+        use cosmwasm_std::{
+            attr, Empty, Event, IbcAcknowledgement, IbcOrder, Reply, ReplyOn,
+            SubMsgExecutionResponse,
+        };
         static CONTRACT: &[u8] = include_bytes!("../testdata/ibc_reflect.wasm");
         const IBC_VERSION: &str = "ibc-reflect-v1";
         fn setup(
@@ -641,16 +710,17 @@ mod tests {
             )
             .unwrap()
             .unwrap();
-            assert_eq!(1, res.submessages.len());
-            let id = res.submessages[0].id;
+            assert_eq!(1, res.messages.len());
+            assert_eq!(ReplyOn::Success, res.messages[0].reply_on);
+            let id = res.messages[0].id;
             let event = Event {
-                kind: "message".into(),
+                ty: "message".into(),
                 attributes: vec![attr("contract_address", &account)],
             };
             // which creates a reflect account. here we get the callback
             let response = Reply {
                 id,
-                result: ContractResult::Ok(SubcallResponse {
+                result: ContractResult::Ok(SubMsgExecutionResponse {
                     events: vec![event],
                     data: None,
                 }),
@@ -678,8 +748,8 @@ mod tests {
             let mut instance = mock_instance(&CONTRACT, &[]);
             setup(&mut instance, CHANNEL_ID, ACCOUNT);
             let packet = mock_ibc_packet_ack(CHANNEL_ID, br#"{}"#).unwrap();
-            let ack = IbcAcknowledgement {
-                acknowledgement: br#"{}"#.into(),
+            let ack = IbcAcknowledgementWithPacket {
+                acknowledgement: IbcAcknowledgement::new(br#"{}"#),
                 original_packet: packet,
             };
             call_ibc_packet_ack::<_, _, _, Empty>(&mut instance, &mock_env(), &ack)
